#include <stdbool.h>
#include <string.h>

#include "lnm/http/loop.h"
#include "lnm/http/loop_internal.h"
#include "lnm/http/req.h"
#include "lnm/loop.h"
#include "lnm/loop_internal.h"

void lnm_http_loop_process_parse_req(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;

  lnm_http_parse_err res = lnm_http_req_parse(
      &ctx->req, &conn->r.buf[conn->r.read], conn->r.size - conn->r.read);

  switch (res) {
  case lnm_http_parse_err_ok:
    conn->r.read += ctx->req.len;
    ctx->state = lnm_http_loop_state_steps;
    break;
  case lnm_http_parse_err_incomplete:
    // If the request is already the size of the read buffer, we close the
    // request. Otherwise, we wait for anything read
    if (conn->r.size - conn->r.read == LNM_LOOP_BUF_SIZE) {
      conn->state = lnm_loop_state_end;
    }
    break;
  case lnm_http_parse_err_invalid:
    conn->state = lnm_loop_state_end;
    break;
  case lnm_http_parse_err_unknown_method:
    // TODO set status code here
    conn->state = lnm_loop_state_end;
    break;
  }
}

void lnm_http_loop_process_route(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;
  lnm_http_loop_gctx *gctx = ctx->g;

  // 0: no match
  // 1: matched route, but not method
  // 2: fully matched route
  int match_level = 0;
  lnm_http_route *route;

  for (size_t i = 0; i < gctx->routes.len && match_level < 2; i++) {
    route = gctx->routes.arr[i];
    bool matched_path;

    switch (route->type) {
    case lnm_http_route_type_literal:
      matched_path =
          strncmp(route->route.s, ctx->req.path.s, ctx->req.path.len) == 0;
      break;
    case lnm_http_route_type_regex:
      matched_path =
          regexec(route->route.regex, ctx->req.path.s,
                  LNM_HTTP_MAX_REGEX_GROUPS, ctx->req.path.groups, 0) == 0;
    }

    // Remember the previous match levels
    int new_match_level = matched_path + (route->method == ctx->req.method);
    match_level = match_level < new_match_level ? new_match_level : match_level;
  }

  ctx->route = match_level == 2 ? route : NULL;
}

void lnm_http_loop_process_steps(lnm_http_conn *conn) {
  /* lnm_http_loop_ctx *ctx = conn->ctx; */

  /* while () */
}

void (*process_fns[])(lnm_http_conn *conn) = {
    lnm_http_loop_process_parse_req,
    lnm_http_loop_process_route,
    lnm_http_loop_process_steps,
};

void lnm_http_loop_process(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;

  lnm_http_loop_state cur_state;

  // As long as the processing is able to advance to a next state, we keep
  // progressing
  do {
    cur_state = ctx->state;

    process_fns[cur_state](conn);
  } while (conn->state == lnm_loop_state_req && cur_state != ctx->state);
}
